# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Class to manage the display of an Update All dialog."""
import threading
from enum import IntEnum, auto
import os
from typing import List, Set

import NetworkManager
from PySideWrapper import QtCore, QtWidgets

import addonmanager_freecad_interface as fci
from Addon import Addon, MissingDependencies
from addonmanager_installer_gui import AddonDependencyInstallerGUI
from addonmanager_installer import AddonInstaller, MacroInstaller


translate = fci.translate

# pylint: disable=too-few-public-methods,too-many-instance-attributes


class UpdaterFactory:
    """A factory class for generating updaters. Mainly exists to allow easily mocking
    those updaters during testing. A replacement class only needs to provide a
    "get_updater" function that returns mock updater objects. Those objects must be
    QObjects with a run() function and a finished signal."""

    def __init__(self, addons):
        self.addons = addons

    def get_updater(self, addon):
        """Get an updater for this addon (either a MacroInstaller or an
        AddonInstaller)"""
        if addon.macro is not None:
            return MacroInstaller(addon)
        return AddonInstaller(addon, self.addons)


class AddonStatus(IntEnum):
    """The current status of the installation process for a given addon"""

    WAITING = auto()
    INSTALLING = auto()
    SUCCEEDED = auto()
    FAILED = auto()

    def ui_string(self):
        """Get the string that the UI should show for this status"""
        if self.value == AddonStatus.WAITING:
            return ""
        if self.value == AddonStatus.INSTALLING:
            return translate("AddonsInstaller", "Installing") + "…"
        if self.value == AddonStatus.SUCCEEDED:
            return translate("AddonsInstaller", "Succeeded")
        if self.value == AddonStatus.FAILED:
            return translate("AddonsInstaller", "Failed")
        return "[INTERNAL ERROR]"


class UpdateAllWorker(QtCore.QObject):
    """A worker to run the "update all" process. Do not run on the main GUI thread."""

    finished = QtCore.Signal()
    addon_updated = QtCore.Signal(object)
    progress_update = QtCore.Signal(int, int)

    def __init__(self, addons: List[Addon]):
        super().__init__()
        self.addons = addons
        self.active_installer = None
        self.updater_factory = UpdaterFactory(addons)
        self.running = False
        self.cancelled = False
        self.currentIndex = 0
        self.download_size_lock = threading.Lock()
        self.total_size = 0
        self.sizes_received = 0
        self.downloaded_sizes = []
        NetworkManager.AM_NETWORK_MANAGER.content_length.connect(self._update_download_size)

    def run(self):
        """Run the Update All process. Blocks until updates are complete or cancelled."""
        self.running = True
        self.currentIndex = 0
        self.query_sizes()

    def query_sizes(self):
        """In the background, builds a list of the download sizes for all the addons being updated"""
        forced_repos = fci.Preferences().get("force_git_in_repos").split(",")
        for addon in self.addons:
            if addon.name in forced_repos:
                self.sizes_received += 1
                continue
            zip_url = addon.get_zip_url()
            NetworkManager.AM_NETWORK_MANAGER.query_download_size(zip_url)

    def _update_download_size(self, _index: int, _response_code: int, content_length: int) -> None:
        with self.download_size_lock:
            self.sizes_received += 1
            self.total_size += content_length
            self.downloaded_sizes.append(0)
            if self.sizes_received == len(self.addons):
                self._process_next_update()

    def cancel(self):
        self.cancelled = True

    def _process_next_update(self):
        """Grab the next addon in the list and start its updater."""
        if not self.cancelled and self.currentIndex < len(self.addons):
            addon = self.addons[self.currentIndex]
            self.currentIndex += 1
            self.active_installer = self.updater_factory.get_updater(addon)
            self._launch_active_installer()
        else:
            self._finalize()

    def _launch_active_installer(self):
        """Set up and run the active installer in a new thread."""

        self.active_installer.success.connect(self._update_succeeded)
        self.active_installer.failure.connect(self._update_failed)
        self.active_installer.finished.connect(self._update_finished)
        if hasattr(self.active_installer, "progress_update"):
            self.active_installer.progress_update.connect(self._update_progress_update)
        self.active_installer.run()

    def _update_progress_update(self, progress: int, total: int) -> None:
        """Calculate total progress and emit the signal"""
        size_array_index = self.currentIndex - 1
        if 0 <= size_array_index < len(self.downloaded_sizes):
            self.downloaded_sizes[size_array_index] = progress
            total_so_far = sum(self.downloaded_sizes)
            self.progress_update.emit(total_so_far, self.total_size)

    def _update_succeeded(self, addon):
        """Callback for a successful update"""
        self.addon_updated.emit(addon)

    def _update_failed(self, addon):
        """Callback for a failed update - does nothing at present"""

    def _update_finished(self):
        """Callback for updater that has finished all its work"""
        if not self.cancelled:
            self._process_next_update()
        else:
            self._setup_cancelled_state()

    def _finalize(self):
        """No more updates, clean up and shut down"""
        self.running = False
        self.finished.emit()

    def _setup_cancelled_state(self):
        self.running = False
        self.finished.emit()

    def is_running(self):
        """True if the thread is running, and False if not"""
        return self.running


class UpdateAllGUI(QtCore.QObject):
    """A GUI to display and manage an "update all" process."""

    finished = QtCore.Signal()
    addon_updated = QtCore.Signal(object)

    def __init__(self, addons: List[Addon]):
        super().__init__()
        self.model = UpdatesAvailableModel(addons)
        self.model.dataChanged.connect(self.model_changed)
        self._setup_main_dialog()
        self._setup_progress_dialog()

        self.addon_installer = None
        self.worker_thread = None
        self.running = False
        self.cancelled = False

        self.dependency_installer = None

    def _setup_main_dialog(self):
        self.dialog = fci.loadUi(os.path.join(os.path.dirname(__file__), "update_all.ui"))
        self.dialog.setObjectName("AddonManager_UpdateAllDialog")
        self.dialog.table_view.setModel(self.model)
        self.dialog.update_button.clicked.connect(self.update_button_clicked)
        self.dialog.button_box.rejected.connect(self.finished.emit)

        self.dialog.table_view.horizontalHeader().setStretchLastSection(False)
        self.dialog.table_view.horizontalHeader().setSectionResizeMode(
            0, QtWidgets.QHeaderView.Stretch
        )
        self.dialog.table_view.horizontalHeader().setSectionResizeMode(
            1, QtWidgets.QHeaderView.ResizeToContents
        )
        self.dialog.table_view.horizontalHeader().setSectionResizeMode(
            2, QtWidgets.QHeaderView.ResizeToContents
        )
        self.dialog.table_view.horizontalHeader().setSectionResizeMode(
            3, QtWidgets.QHeaderView.ResizeToContents
        )
        self.dialog.table_view.horizontalHeader().setSectionResizeMode(
            4, QtWidgets.QHeaderView.ResizeToContents
        )
        self.dialog.table_view.hideColumn(4)

    def _setup_progress_dialog(self):
        self.progress_dialog = fci.loadUi(os.path.join(os.path.dirname(__file__), "progress.ui"))
        self.progress_dialog.setObjectName("AddonManager_UpdateAllProgressDialog")
        self.progress_dialog.buttonBox.rejected.connect(self.cancel)

    def run(self):
        """Runs the update selection modal dialog."""
        self.running = True
        self.dialog.show()

    def update_button_clicked(self):
        """Runs the updater on all the selected addons. First checks to see if there are any
        dependencies that need to be installed. If so, it prompts the user to confirm that they
        want to install them."""

        missing_deps = MissingDependencies()

        addons_selected = []
        for checked, addon in zip(self.model.update_is_checked, self.model.addons_with_update):
            if checked:
                fci.Console.PrintMessage(f"Preparing to update {addon.display_name}\n")
                addons_selected.append(addon)
                missing_deps.import_from_addon(addon, self.model.addons)

        required_wbs = set(missing_deps.wbs)
        required_addons = set(missing_deps.external_addons)
        required_python_modules = set(missing_deps.python_requires)
        optional_python_modules = set(missing_deps.python_optional)

        if required_wbs or required_addons or required_python_modules or optional_python_modules:
            fci.Console.PrintMessage(
                "Found unsatisfied dependencies for the requested addon updates\n"
            )
            if required_wbs:
                fci.Console.PrintMessage(f"  Required Workbenches: {required_wbs}\n")
            if required_addons:
                fci.Console.PrintMessage(
                    f"  Required Addons: {','.join([x.display_name for x in required_addons])}\n"
                )
            if required_python_modules:
                fci.Console.PrintMessage(f"  Required Python Modules: {required_python_modules}\n")
            if optional_python_modules:
                fci.Console.PrintMessage(f"  Optional Python Modules: {optional_python_modules}\n")
            self.handle_missing_dependencies(
                addons_selected,
                required_wbs,
                required_addons,
                required_python_modules,
                optional_python_modules,
            )
        else:
            fci.Console.PrintMessage("No unsatisfied dependencies found, continuing with update\n")
            self.check_for_git_migration()

    def handle_missing_dependencies(
        self,
        addons: List[Addon],
        required_wbs: Set[str],
        required_addons: Set[Addon],
        required_python_modules: Set[str],
        optional_python_modules: Set[str],
    ):
        missing_dependencies = MissingDependencies()
        missing_dependencies.wbs = list(required_wbs)
        missing_dependencies.external_addons = list(required_addons)
        missing_dependencies.python_requires = list(required_python_modules)
        missing_dependencies.python_optional = list(optional_python_modules)
        self.dependency_installer = AddonDependencyInstallerGUI(addons, missing_dependencies)
        self.dependency_installer.cancel.connect(self.cancel)
        self.dependency_installer.proceed.connect(self.check_for_git_migration)
        self.dependency_installer.run()

    def check_for_git_migration(self):
        """The Addon Manager used to use git as the preferred installation mechanism and only
        fell back to Zip if git was unavailable (or specifically deactivated). That changed in
        mid-2025, and Zip was made the preferred download mechanism. This function checks to see if
        there is a .git directory present that would be deleted by doing this update and offers
        to let the user turn on git installation (if they are a developer, for example)."""

        addons_to_update = [
            addon
            for addon, checked in zip(self.model.addons_with_update, self.model.update_is_checked)
            if checked
        ]
        custom_repos_lines = fci.Preferences().get("CustomRepositories").split("\n")
        custom_repos = [line.split(" ")[0] for line in custom_repos_lines]
        forced_repos = fci.Preferences().get("force_git_in_repos").split(",")
        for addon in addons_to_update:
            if addon.name in custom_repos or addon.name in forced_repos:
                continue
            path_to_addon = str(os.path.join(fci.DataPaths().mod_dir, addon.name))
            path_to_git_directory = str(os.path.join(path_to_addon, ".git"))
            if os.path.exists(path_to_git_directory):
                backup_path = path_to_addon + "-backup-before-zip-migration"
                os.rename(path_to_addon, backup_path)
                fci.Console.PrintMessage(
                    f"Found .git directory for Addon {addon.display_name}"
                    " - backup created before migration to zip\n"
                )
                with open(os.path.join(backup_path, "ADDON_DISABLED"), "w", encoding="utf-8") as f:
                    f.write(
                        "This directory is a backup made before migrating from Git to Zip."
                        " If you don't care about retaining any git information or you are"
                        " not a developer and/or don't use git, it can be deleted safely."
                    )
        self.proceed()

    def proceed(self):
        """Does the updates"""

        if self.worker_thread is not None and self.worker_thread.isRunning():
            self.worker_thread.quit()
            self.worker_thread.wait()

        self.progress_dialog.show()
        self.dialog.table_view.showColumn(4)  # The "Done" checkbox column
        addons_to_update = [
            addon
            for addon, checked in zip(self.model.addons_with_update, self.model.update_is_checked)
            if checked
        ]
        self.progress_dialog.progressBar.setMaximum(len(addons_to_update))
        self.progress_dialog.progressBar.setValue(0)
        self.addon_installer = UpdateAllWorker(addons_to_update)
        self.worker_thread = QtCore.QThread()
        self.addon_installer.moveToThread(self.worker_thread)
        self.worker_thread.started.connect(self.addon_installer.run)
        self.addon_installer.finished.connect(self.worker_thread.quit)
        self.addon_installer.finished.connect(self.update_complete)
        self.addon_installer.addon_updated.connect(self.update_progress)
        self.addon_installer.progress_update.connect(self.update_progress_bar)
        self.worker_thread.start()

    def cancel(self):
        """Cancels the running updates but leaves the window open and allows the user to try again
        if they want to."""
        self.cancelled = True
        if self.addon_installer is not None:
            self.addon_installer.cancel()
        if self.progress_dialog is not None:
            self.progress_dialog.hide()
        if not any(self.model.update_is_checked):
            self.dialog.update_button.setEnabled(False)
        else:
            self.dialog.update_button.setEnabled(True)

    def update_progress(self, addon_installed: Addon):
        """Updates the progress bar and check the state of the rows"""
        self.model.rescan_addon(addon_installed)
        self.addon_updated.emit(addon_installed)

    def update_progress_bar(self, so_far: int, total: int):
        self.progress_dialog.progressBar.setMaximum(total)
        self.progress_dialog.progressBar.setValue(so_far)

    def update_complete(self):
        self.progress_dialog.hide()

    def model_changed(
        self,
        _top_left: QtCore.QModelIndex,
        _bottom_right: QtCore.QModelIndex,
        _roles: List[QtCore.Qt.ItemDataRole],
    ):
        if not any(self.model.update_is_checked):
            self.dialog.update_button.setEnabled(False)
        else:
            self.dialog.update_button.setEnabled(True)


class UpdatesAvailableModel(QtCore.QAbstractTableModel):
    """A model to manage the list of updates available"""

    def __init__(self, addons: List[Addon]):
        super().__init__()
        self.addons = addons
        self.addons_with_update: List[Addon] = []
        self.update_is_checked: List[bool] = []
        self.update_is_done: List[bool] = []
        self.headers = [
            translate("AddonsInstaller", "Name", "Column header"),
            translate("AddonsInstaller", "Installed Version", "Column header"),
            translate("AddonsInstaller", "Available Version", "Column header"),
            translate("AddonsInstaller", "Update?", "Column header"),
            translate("AddonsInstaller", "Done", "Column header"),
        ]
        self.setup()

    def setup(self):
        for addon in self.addons:
            if addon.status() == Addon.Status.UPDATE_AVAILABLE:
                self.addons_with_update.append(addon)
                self.update_is_checked.append(True)
                self.update_is_done.append(False)

    def rescan_addon(self, addon: Addon):
        for i, a in enumerate(self.addons_with_update):
            if a == addon:
                self.update_is_done[i] = True
                self.update_is_checked[i] = False
                self.dataChanged.emit(
                    self.createIndex(i, 1),
                    self.createIndex(i, 4),
                    [QtCore.Qt.DisplayRole, QtCore.Qt.CheckStateRole],
                )
                return

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid():
            return 0
        return len(self.addons_with_update)

    def columnCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid():
            return 0
        return len(self.headers)

    def data(self, index: QtCore.QModelIndex, role: int = QtCore.Qt.DisplayRole):
        if not index.isValid():
            return None
        if role == QtCore.Qt.DisplayRole:
            addon = self.addons_with_update[index.row()]
            if index.column() == 0:
                return addon.display_name
            if index.column() == 1:
                if self.update_is_done[index.row()]:
                    # Make sure that this column matches the next one if we just updated, otherwise
                    # it's very confusing!
                    return str(addon.metadata.version) if addon.metadata else ""
                return str(addon.installed_metadata.version) if addon.installed_metadata else ""
            if index.column() == 2:
                return str(addon.metadata.version) if addon.metadata else ""
        elif role == QtCore.Qt.CheckStateRole:
            if index.column() == 3:
                if self.update_is_done[index.row()]:
                    return QtCore.Qt.Unchecked
                return (
                    QtCore.Qt.Checked
                    if self.update_is_checked[index.row()]
                    else QtCore.Qt.Unchecked
                )
            if index.column() == 4:
                return (
                    QtCore.Qt.Checked if self.update_is_done[index.row()] else QtCore.Qt.Unchecked
                )
        return None

    def flags(self, index: QtCore.QModelIndex):
        if not index.isValid():
            return QtCore.Qt.NoItemFlags

        if index.column() == 3:
            if self.update_is_done[index.row()]:
                return QtCore.Qt.NoItemFlags
            return (
                QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEditable
            )

        return QtCore.Qt.NoItemFlags

    def setData(self, index: QtCore.QModelIndex, value, role: int = QtCore.Qt.EditRole):
        if not index.isValid():
            return False

        if index.column() == 3 and role == QtCore.Qt.CheckStateRole:
            self.update_is_checked[index.row()] = QtCore.Qt.CheckState(value) == QtCore.Qt.Checked
            self.dataChanged.emit(index, index, [QtCore.Qt.CheckStateRole])
            return True

        return False

    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.DisplayRole and orientation == QtCore.Qt.Horizontal:
            return self.headers[section]
        return None



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import Path.Tool
from PySide.QtCore import QT_TRANSLATE_NOOP
from ...toolbit import ToolBit
from ...assets.ui import AssetSaveDialog
from ..serializers import all_serializers as toolbit_serializers
from .file import ToolBitOpenDialog

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class CommandToolBitCreate:
    """
    Command used to create a new Tool.
    """

    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "CAM_ToolBit",
            "MenuText": QT_TRANSLATE_NOOP("CAM_ToolBitCreate", "New Toolbit"),
            "ToolTip": QT_TRANSLATE_NOOP("CAM_ToolBitCreate", "Creates a new toolbit object"),
        }

    def IsActive(self):
        return FreeCAD.ActiveDocument is not None

    def Activated(self):
        # Create a default endmill tool bit and attach it to a new DocumentObject
        toolbit = ToolBit.from_shape_id("endmill.fcstd")
        obj = toolbit.attach_to_doc(FreeCAD.ActiveDocument)
        obj.ViewObject.Proxy.setCreate(obj.ViewObject)


class CommandToolBitSave:
    """
    Command used to save an existing Tool to a file.
    """

    def __init__(self, saveAs):
        self.saveAs = saveAs

    def GetResources(self):
        if self.saveAs:
            menuTxt = QT_TRANSLATE_NOOP("CAM_ToolBitSaveAs", "Save Tool As…")
        else:
            menuTxt = QT_TRANSLATE_NOOP("CAM_ToolBitSave", "Save Tool")
        return {
            "Pixmap": "CAM_ToolBit",
            "MenuText": menuTxt,
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_ToolBitSave", "Saves an existing toolbit object to a file"
            ),
        }

    def selectedTool(self):
        sel = FreeCADGui.Selection.getSelectionEx()
        if 1 == len(sel) and isinstance(sel[0].Object.Proxy, Path.Tool.ToolBit):
            return sel[0].Object
        return None

    def IsActive(self):
        tool = self.selectedTool()
        if tool:
            if tool.File:
                return True
            return self.saveAs
        return False

    def Activated(self):
        tool_obj = self.selectedTool()
        if not tool_obj:
            return
        toolbit = tool_obj.Proxy

        dialog = AssetSaveDialog(ToolBit, toolbit_serializers, FreeCADGui.getMainWindow())
        dialog_result = dialog.exec_(toolbit)
        if not dialog_result:
            return


class CommandToolBitLoad:
    """
    Command used to load an existing Tool from a file into the current document.
    """

    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "CAM_ToolBit",
            "MenuText": QT_TRANSLATE_NOOP("CAM_ToolBitLoad", "Load Tool"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_ToolBitLoad", "Loads an existing toolbit object from a file"
            ),
        }

    def selectedTool(self):
        sel = FreeCADGui.Selection.getSelectionEx()
        if 1 == len(sel) and isinstance(sel[0].Object.Proxy, Path.Tool.ToolBit):
            return sel[0].Object
        return None

    def IsActive(self):
        return FreeCAD.ActiveDocument is not None

    def Activated(self):
        dialog = ToolBitOpenDialog(toolbit_serializers, FreeCADGui.getMainWindow())
        toolbits = dialog.exec()
        for toolbit in toolbits:
            toolbit.attach_to_doc(FreeCAD.ActiveDocument)
        if toolbits:
            FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("CAM_ToolBitCreate", CommandToolBitCreate())
    FreeCADGui.addCommand("CAM_ToolBitLoad", CommandToolBitLoad())
    FreeCADGui.addCommand("CAM_ToolBitSave", CommandToolBitSave(False))
    FreeCADGui.addCommand("CAM_ToolBitSaveAs", CommandToolBitSave(True))

CommandList = [
    "CAM_ToolBitCreate",
    "CAM_ToolBitLoad",
    "CAM_ToolBitSave",
    "CAM_ToolBitSaveAs",
]



# SPDX-License-Identifier: LGPL-2.1-or-later

import yaml
import json
from typing import Type, cast
import FreeCAD
from CAMTests.PathTestUtils import PathTestWithAssets
from Path.Tool.toolbit import ToolBit, ToolBitEndmill
from Path.Tool.toolbit.serializers import (
    FCTBSerializer,
    CamoticsToolBitSerializer,
    YamlToolBitSerializer,
)
from Path.Tool.assets.asset import Asset
from Path.Tool.assets.serializer import AssetSerializer
from Path.Tool.assets.uri import AssetUri
from Path.Tool.shape import ToolBitShapeEndmill
from typing import Mapping


class _BaseToolBitSerializerTestCase(PathTestWithAssets):
    """Base test case for ToolBit Serializers."""

    __test__ = False

    serializer_class: Type[AssetSerializer]
    test_tool_bit: ToolBit

    def setUp(self):
        """Create a tool bit for each test."""
        super().setUp()
        if self.serializer_class is None or not issubclass(self.serializer_class, AssetSerializer):
            raise NotImplementedError("Subclasses must define a valid serializer_class")

        self.test_tool_bit = cast(ToolBitEndmill, self.assets.get("toolbit://5mm_Endmill"))
        self.test_tool_bit.label = "Test Tool"
        self.test_tool_bit.set_diameter(FreeCAD.Units.Quantity("4.12 mm"))
        self.test_tool_bit.set_length(FreeCAD.Units.Quantity("15.0 mm"))

    def test_serialize(self):
        """Test serialization of a toolbit."""
        if self.test_tool_bit is None:
            raise NotImplementedError("Subclasses must define a test_tool_bit")
        serialized_data = self.serializer_class.serialize(self.test_tool_bit)
        self.assertIsInstance(serialized_data, bytes)

    def test_extract_dependencies(self):
        """Test dependency extraction."""
        # This test assumes that the serializers don't have dependencies
        # and can be overridden in subclasses if needed.
        serialized_data = self.serializer_class.serialize(self.test_tool_bit)
        dependencies = self.serializer_class.extract_dependencies(serialized_data)
        self.assertIsInstance(dependencies, list)
        self.assertEqual(len(dependencies), 0)


class TestCamoticsToolBitSerializer(_BaseToolBitSerializerTestCase):
    serializer_class = CamoticsToolBitSerializer

    def test_serialize(self):
        super().test_serialize()
        serialized_data = self.serializer_class.serialize(self.test_tool_bit)
        # Camotics specific assertions
        expected_substrings = [
            b'"units": "metric"',
            b'"shape": "Cylindrical"',
            b'"length": 15',
            b'"diameter": 4.12',
            b'"description": "Test Tool"',
        ]
        for substring in expected_substrings:
            self.assertIn(substring, serialized_data)

    def test_deserialize(self):
        # Create a known serialized data string based on the Camotics format
        camotics_data = (
            b'{"units": "metric", "shape": "Cylindrical", "length": 15, '
            b'"diameter": 4.12, "description": "Test Tool"}'
        )
        deserialized_bit = cast(
            ToolBitEndmill,
            self.serializer_class.deserialize(camotics_data, id="test_id", dependencies=None),
        )

        self.assertIsInstance(deserialized_bit, ToolBit)
        self.assertEqual(deserialized_bit.label, "Test Tool")
        self.assertEqual(
            deserialized_bit.get_diameter(), FreeCAD.Units.Quantity(4.12, FreeCAD.Units.Length)
        )
        self.assertEqual(
            deserialized_bit.get_length(), FreeCAD.Units.Quantity(15.0, FreeCAD.Units.Length)
        )
        self.assertEqual(deserialized_bit.get_shape_name(), "Endmill")


class TestFCTBSerializer(_BaseToolBitSerializerTestCase):
    serializer_class = FCTBSerializer

    def test_serialize(self):
        super().test_serialize()
        serialized_data = self.serializer_class.serialize(self.test_tool_bit)
        # FCTB specific assertions (JSON format)
        data = json.loads(serialized_data.decode("utf-8"))
        self.assertEqual(data.get("name"), "Test Tool")
        self.assertEqual(data.get("shape"), "endmill.fcstd")
        self.assertEqual(
            FreeCAD.Units.Quantity(data.get("parameter", {}).get("Diameter")),
            FreeCAD.Units.Quantity(4.12, FreeCAD.Units.Length),
        )
        self.assertEqual(
            FreeCAD.Units.Quantity(data.get("parameter", {}).get("Length")),
            FreeCAD.Units.Quantity(15.0, FreeCAD.Units.Length),
        )

    def test_extract_dependencies(self):
        """Test dependency extraction for FCTB."""
        fctb_data = (
            b'{"name": "Test Tool", "pocket": null, "shape": "endmill", '
            b'"parameter": {"Diameter": "4.12 mm", "Length": "15.0 mm"}, "attribute": {}}'
        )
        dependencies = self.serializer_class.extract_dependencies(fctb_data)
        self.assertIsInstance(dependencies, list)
        self.assertEqual(len(dependencies), 1)
        self.assertEqual(dependencies[0], AssetUri.build("toolbitshape", "endmill"))

    def test_deserialize(self):
        # Create a known serialized data string based on the FCTB format
        fctb_data = (
            b'{"name": "Test Tool", "pocket": null, "shape": "endmill", '
            b'"parameter": {"Diameter": "4.12 mm", "Length": "15.0 mm"}, "attribute": {}}'
        )
        # Create a ToolBitShapeEndmill instance for 'endmill'
        shape = ToolBitShapeEndmill("endmill")

        # Create the dependencies dictionary with the shape instance
        dependencies: Mapping[AssetUri, Asset] = {AssetUri.build("toolbitshape", "endmill"): shape}

        # Provide dummy id and dependencies for deserialization test
        deserialized_bit = cast(
            ToolBitEndmill,
            self.serializer_class.deserialize(fctb_data, id="test_id", dependencies=dependencies),
        )

        self.assertIsInstance(deserialized_bit, ToolBit)
        self.assertEqual(deserialized_bit.label, "Test Tool")
        self.assertEqual(deserialized_bit.get_shape_name(), "Endmill")
        self.assertEqual(
            deserialized_bit.get_diameter(), FreeCAD.Units.Quantity(4.12, FreeCAD.Units.Length)
        )
        self.assertEqual(
            deserialized_bit.get_length(), FreeCAD.Units.Quantity(15.0, FreeCAD.Units.Length)
        )


class TestYamlToolBitSerializer(_BaseToolBitSerializerTestCase):
    serializer_class = YamlToolBitSerializer

    def test_serialize(self):
        super().test_serialize()
        serialized_data = self.serializer_class.serialize(self.test_tool_bit)
        # YAML specific assertions
        data = yaml.safe_load(serialized_data.decode("utf-8"))
        self.assertEqual(data.get("id"), "5mm_Endmill")
        self.assertEqual(data.get("name"), "Test Tool")
        self.assertEqual(data.get("shape"), "endmill.fcstd")
        self.assertEqual(data.get("shape-type"), "Endmill")
        self.assertEqual(
            FreeCAD.Units.Quantity(data.get("parameter", {}).get("Diameter")),
            FreeCAD.Units.Quantity(4.12, FreeCAD.Units.Length),
        )
        self.assertEqual(
            FreeCAD.Units.Quantity(data.get("parameter", {}).get("Length")),
            FreeCAD.Units.Quantity(15.0, FreeCAD.Units.Length),
        )

    def test_extract_dependencies(self):
        """Test dependency extraction for YAML."""
        yaml_data = (
            b"name: Test Tool\n"
            b"shape: endmill\n"
            b"shape-type: Endmill\n"
            b"parameter:\n"
            b"  Diameter: 4.12 mm\n"
            b"  Length: 15.0 mm\n"
            b"attribute: {}\n"
        )
        dependencies = self.serializer_class.extract_dependencies(yaml_data)
        self.assertIsInstance(dependencies, list)
        self.assertEqual(len(dependencies), 1)
        self.assertEqual(dependencies[0], AssetUri.build("toolbitshape", "endmill"))

    def test_deserialize(self):
        # Create a known serialized data string based on the YAML format
        yaml_data = (
            b"id: TestID\n"
            b"name: Test Tool\n"
            b"shape: endmill\n"
            b"shape-type: Endmill\n"
            b"parameter:\n"
            b"  Diameter: 4.12 mm\n"
            b"  Length: 15.0 mm\n"
            b"attribute: {}\n"
        )
        # Create a ToolBitShapeEndmill instance for 'endmill'
        shape = ToolBitShapeEndmill("endmill")

        # Create the dependencies dictionary with the shape instance
        dependencies: Mapping[AssetUri, Asset] = {AssetUri.build("toolbitshape", "endmill"): shape}

        # Provide dummy id and dependencies for deserialization test
        deserialized_bit = cast(
            ToolBitEndmill,
            self.serializer_class.deserialize(yaml_data, "TestID", dependencies=dependencies),
        )
        self.assertIsInstance(deserialized_bit, ToolBit)
        self.assertEqual(deserialized_bit.id, "TestID")
        self.assertEqual(deserialized_bit.label, "Test Tool")
        self.assertEqual(deserialized_bit.get_shape_name(), "Endmill")
        self.assertEqual(
            deserialized_bit.get_diameter(), FreeCAD.Units.Quantity(4.12, FreeCAD.Units.Length)
        )
        self.assertEqual(
            deserialized_bit.get_length(), FreeCAD.Units.Quantity(15.0, FreeCAD.Units.Length)
        )

        # Test with ID argument.
        deserialized_bit = cast(
            ToolBitEndmill,
            self.serializer_class.deserialize(yaml_data, id="test_id", dependencies=dependencies),
        )
        self.assertEqual(deserialized_bit.id, "test_id")



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint tie document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package constraint_tie
#  \ingroup FEM
#  \brief constraint tie object

from . import base_fempythonobject

import FreeCAD

_PropHelper = base_fempythonobject._PropHelper


class ConstraintTie(base_fempythonobject.BaseFemPythonObject):
    """
    The ConstraintTie object
    """

    Type = "Fem::ConstraintTie"

    def __init__(self, obj):
        super().__init__(obj)

        for prop in self._get_properties():
            prop.add_to_object(obj)

    def _get_properties(self):
        prop = []

        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="Tolerance",
                group="Geometry",
                doc="Set max gap between tied faces",
                value="0.0 mm",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="Adjust",
                group="Geometry",
                doc="Adjust connected nodes",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyBool",
                name="CyclicSymmetry",
                group="Geometry",
                doc="Define cyclic symmetry model",
                value=False,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyPlacement",
                name="SymmetryAxis",
                group="Geometry",
                doc="Placement of axis of symmetry",
                value=FreeCAD.Placement(),
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyInteger",
                name="Sectors",
                group="Geometry",
                doc="Number of sectors",
                value=0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyInteger",
                name="ConnectedSectors",
                group="Geometry",
                doc="Number of connected sectors",
                value=1,
            )
        )

        return prop

    def onDocumentRestored(self, obj):
        # update old project with new properties
        for prop in self._get_properties():
            try:
                obj.getPropertyByName(prop.name)
            except FreeCAD.Base.PropertyError:
                prop.add_to_object(obj)



# ***************************************************************************
# *   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import unittest

import FreeCAD
import Part
import Sketcher
import TestSketcherApp

App = FreeCAD


class TestPipe(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestPipe")

    def testSimpleAdditivePipeCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.ProfileSketch = self.Doc.addObject("Sketcher::SketchObject", "ProfileSketch")
        self.Body.addObject(self.ProfileSketch)
        TestSketcherApp.CreateCircleSketch(self.ProfileSketch, (0, 0), 1)
        self.Doc.recompute()
        self.SpineSketch = self.Doc.addObject("Sketcher::SketchObject", "SpineSketch")
        self.Body.addObject(self.SpineSketch)
        self.SpineSketch.MapMode = "FlatFace"
        self.SpineSketch.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        self.Doc.recompute()
        self.SpineSketch.addGeometry(
            Part.LineSegment(App.Vector(0.0, 0.0, 0), App.Vector(0, 1, 0)), False
        )
        self.SpineSketch.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -1, 1))
        self.SpineSketch.addConstraint(Sketcher.Constraint("PointOnObject", 0, 2, -2))
        self.SpineSketch.addConstraint(Sketcher.Constraint("DistanceY", 0, 1, 0, 2, 1))
        self.Doc.recompute()
        self.AdditivePipe = self.Doc.addObject("PartDesign::AdditivePipe", "AdditivePipe")
        self.Body.addObject(self.AdditivePipe)
        self.AdditivePipe.Profile = self.ProfileSketch
        self.AdditivePipe.Spine = self.SpineSketch
        self.Doc.recompute()
        self.assertAlmostEqual(self.AdditivePipe.Shape.Volume, 3.14159265)

    def testSimpleSubtractivePipeCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.ProfileSketch = self.Doc.addObject("Sketcher::SketchObject", "ProfileSketch")
        self.Body.addObject(self.ProfileSketch)
        TestSketcherApp.CreateCircleSketch(self.ProfileSketch, (0, 0), 1)
        self.Doc.recompute()
        self.SpineSketch = self.Doc.addObject("Sketcher::SketchObject", "SpineSketch")
        self.Body.addObject(self.SpineSketch)
        self.SpineSketch.MapMode = "FlatFace"
        self.SpineSketch.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        self.Doc.recompute()
        self.SpineSketch.addGeometry(
            Part.LineSegment(App.Vector(0.0, 0.0, 0), App.Vector(0, 1, 0)), False
        )
        self.SpineSketch.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -1, 1))
        self.SpineSketch.addConstraint(Sketcher.Constraint("PointOnObject", 0, 2, -2))
        self.SpineSketch.addConstraint(Sketcher.Constraint("DistanceY", 0, 1, 0, 2, 1))
        self.Doc.recompute()
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "PadSketch")
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateRectangleSketch(self.PadSketch, (-5, -5), (10, 10))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Body.addObject(self.Pad)
        self.Pad.Profile = self.PadSketch
        self.Pad.Length = 1.0
        self.Doc.recompute()
        self.SubtractivePipe = self.Doc.addObject("PartDesign::SubtractivePipe", "SubtractivePipe")
        self.Body.addObject(self.SubtractivePipe)
        self.SubtractivePipe.Profile = self.ProfileSketch
        self.SubtractivePipe.Spine = self.SpineSketch
        self.Doc.recompute()
        self.assertAlmostEqual(self.SubtractivePipe.Shape.Volume, 100 - 3.14159265)

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestPipe")
        # print ("omit closing document for debugging")



# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import sys
import FreeCAD
from FreeCAD import Vector

import Draft
import ObjectsFem
import Part

from BOPTools import SplitFeatures
from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Inductive heating - Elmer 2D",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["current density"],
        "solvers": ["elmer"],
        "material": "solid",
        "equations": ["electromagnetic"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.equation_magnetodynamics_2D_elmer import setup
setup()

Magnetodynamic2D equation - Elmer solver

"""
    )


def setup(doc=None, solvertype="elmer"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects

    # wire defining the insulation area
    p1 = Vector(0.0, 0.0, 0.0)
    p2 = Vector(40.0, 0.0, 0.0)
    p3 = Vector(40.0, 120.0, 0.0)
    p4 = Vector(0.0, 120.0, 0.0)
    p5 = Vector(0.0, 100.0, 0.0)
    p6 = Vector(25.0, 100.0, 0.0)
    p7 = Vector(25.0, 20.0, 0.0)
    p8 = Vector(0.0, 20.0, 0.0)
    Insulation = Draft.make_wire([p1, p2, p3, p4, p5, p6, p7, p8], closed=True)
    Insulation.MakeFace = True
    Insulation.Label = "Insulation"
    Insulation.ViewObject.Visibility = False

    # wire defining the coil volume
    p1 = Vector(50.0, -10.0, 0.0)
    p2 = Vector(55.0, -10.0, 0.0)
    p3 = Vector(55.0, 110.0, 0.0)
    p4 = Vector(50.0, 110.0, 0.0)
    Coil = Draft.make_wire([p1, p2, p3, p4], closed=True)
    Coil.MakeFace = True
    Coil.Label = "Coil"
    Coil.ViewObject.Visibility = False

    # wire defining the crucible area
    p1 = Vector(0.0, 20.0, 0.0)
    p2 = Vector(25.0, 20.0, 0.0)
    p3 = Vector(25.0, 100.0, 0.0)
    p4 = Vector(0.0, 100.0, 0.0)
    p5 = Vector(0.0, 90.0, 0.0)
    p6 = Vector(20.0, 90.0, 0.0)
    p7 = Vector(20.0, 30.0, 0.0)
    p8 = Vector(0.0, 30.0, 0.0)
    Crucible = Draft.make_wire([p1, p2, p3, p4, p5, p6, p7, p8], closed=True)
    Crucible.MakeFace = True
    Crucible.Label = "Crucible"
    Crucible.ViewObject.Visibility = False

    # wire defining the powder volume
    p1 = Vector(0.0, 30.0, 0.0)
    p2 = Vector(20.0, 30.0, 0.0)
    p3 = Vector(20.0, 40.0, 0.0)
    p4 = Vector(0.0, 40.0, 0.0)
    Powder = Draft.make_wire([p1, p2, p3, p4], closed=True)
    Powder.MakeFace = True
    Powder.Label = "Powder"
    Powder.ViewObject.Visibility = False

    # a half circle defining later the air volume
    Air_Circle = Part.makeCircle(140.0, Vector(0.0, 60.0, 0.0), Vector(0.0, 0.0, 1.0), -90.0, 90.0)
    Air_Line = Part.makeLine((0.0, -80.0, 0.0), (0.0, 200.0, 0.0))
    Air_Area = doc.addObject("Part::Feature", "Air_Area")
    Air_Area.Shape = Part.Face([Part.Wire([Air_Circle, Air_Line])])

    # a link of the Insulation
    InsulationLink = doc.addObject("App::Link", "Link_Insulation")
    InsulationLink.LinkTransform = True
    InsulationLink.LinkedObject = Insulation
    InsulationLink.ViewObject.Visibility = False

    # a link of the Coil
    CoilLink = doc.addObject("App::Link", "Link_Coil")
    CoilLink.LinkTransform = True
    CoilLink.LinkedObject = Coil
    CoilLink.ViewObject.Visibility = False

    # a link of the Crucible
    CrucibleLink = doc.addObject("App::Link", "Link_Crucible")
    CrucibleLink.LinkTransform = True
    CrucibleLink.LinkedObject = Crucible
    CrucibleLink.ViewObject.Visibility = False

    # a link of the Powder
    PowderLink = doc.addObject("App::Link", "Link_Powder")
    PowderLink.LinkTransform = True
    PowderLink.LinkedObject = Powder
    PowderLink.ViewObject.Visibility = False

    # fusion of all links
    Fusion = doc.addObject("Part::MultiFuse", "Fusion")
    Fusion.Shapes = [InsulationLink, CoilLink, CrucibleLink, PowderLink]
    Fusion.ViewObject.Visibility = False

    # cut all objects from Air wire to get volume of fluid
    Cut = doc.addObject("Part::Cut", "Cut_Air")
    Cut.Base = Air_Area
    Cut.Tool = Fusion
    Cut.ViewObject.Visibility = False

    # BooleanFregments object to combine cut with rod
    BooleanFragments = SplitFeatures.makeBooleanFragments(name="BooleanFragments")
    BooleanFragments.Objects = [Insulation, Coil, Crucible, Powder, Cut]

    # set view
    doc.recompute()
    if FreeCAD.GuiUp:
        BooleanFragments.ViewObject.Document.activeView().viewTop()
        BooleanFragments.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")
    if FreeCAD.GuiUp:
        import FemGui

        FemGui.setActiveAnalysis(analysis)

    # solver
    if solvertype == "elmer":
        solver_obj = ObjectsFem.makeSolverElmer(doc, "SolverElmer")
        solver_obj.CoordinateSystem = "Axi Symmetric"
        equation_magnetodynamic2D = ObjectsFem.makeEquationMagnetodynamic2D(doc, solver_obj)
        equation_magnetodynamic2D.AngularFrequency = "50 kHz"
        equation_magnetodynamic2D.CalculateCurrentDensity = True
        equation_magnetodynamic2D.CalculateElectricField = True
        equation_magnetodynamic2D.CalculateJouleHeating = True
        equation_magnetodynamic2D.IsHarmonic = True
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
        return doc
    analysis.addObject(solver_obj)

    # materials

    # air around the objects and inside the coil
    material_obj = ObjectsFem.makeMaterialFluid(doc, "Air")
    mat = material_obj.Material
    mat["Name"] = "Air"
    mat["Density"] = "1.204 kg/m^3"
    mat["ThermalConductivity"] = "0.02587 W/m/K"
    mat["ThermalExpansionCoefficient"] = "3.43e-3 1/K"
    mat["SpecificHeat"] = "1.01 kJ/kg/K"
    mat["ElectricalConductivity"] = "1e-12 S/m"
    mat["RelativePermeability"] = "1.0"
    mat["RelativePermittivity"] = "1.00059"
    material_obj.Material = mat
    material_obj.References = [
        (BooleanFragments, "Face2"),
        (BooleanFragments, "Face5"),
        (BooleanFragments, "Face6"),
    ]
    analysis.addObject(material_obj)

    # graphite of the crucible
    material_obj = ObjectsFem.makeMaterialSolid(doc, "Material-Crucible")
    mat = material_obj.Material
    mat["Name"] = "Graphite"
    mat["ElectricalConductivity"] = "2e4 S/m"
    mat["RelativePermeability"] = "1.0"
    material_obj.Material = mat
    material_obj.References = [(BooleanFragments, "Face3")]
    analysis.addObject(material_obj)

    # insulation of the crucible
    material_obj = ObjectsFem.makeMaterialSolid(doc, "Material-Insulation")
    mat = material_obj.Material
    mat["Name"] = "Insulation"
    mat["ElectricalConductivity"] = "2.0e3 S/m"
    mat["RelativePermeability"] = "1.0"
    material_obj.Material = mat
    material_obj.References = [(BooleanFragments, "Face1")]
    analysis.addObject(material_obj)

    # material of powder
    material_obj = ObjectsFem.makeMaterialSolid(doc, "Material-Powder")
    mat = material_obj.Material
    mat["Name"] = "Powder"
    mat["ElectricalConductivity"] = "1e4 S/m"
    mat["RelativePermeability"] = "1.0"
    material_obj.Material = mat
    material_obj.References = [(BooleanFragments, "Face4")]
    analysis.addObject(material_obj)

    # constraint current density
    CurrentDensity = ObjectsFem.makeConstraintCurrentDensity(doc, "CurrentDensity")
    CurrentDensity.References = [(BooleanFragments, "Face2")]
    CurrentDensity.NormalCurrentDensity_re = "250000.000 A/m^2"
    CurrentDensity.Mode = "Normal"
    analysis.addObject(CurrentDensity)

    # mesh
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.Shape = BooleanFragments
    femmesh_obj.CharacteristicLengthMax = "3 mm"
    femmesh_obj.ViewObject.Visibility = False

    # mesh_region
    mesh_region = ObjectsFem.makeMeshRegion(doc, femmesh_obj, name="MeshRegion")
    mesh_region.CharacteristicLength = "1 mm"
    mesh_region.References = [
        (BooleanFragments, "Face1"),
        (BooleanFragments, "Face2"),
        (BooleanFragments, "Face3"),
        (BooleanFragments, "Face4"),
    ]
    mesh_region.ViewObject.Visibility = False

    # generate the mesh
    generate_mesh.mesh_from_mesher(femmesh_obj, "gmsh")

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# (c) 2024 Werner Mayer LGPL

__author__ = "Werner Mayer"
__url__ = "https://www.freecad.org"
__doc__ = "Helper module to convert zip files"


import zipfile

def rewrite(source: str, target: str):
    source_zip = zipfile.ZipFile(source, "r")
    target_zip = zipfile.ZipFile(target, "w")

    for name in source_zip.namelist():
        target_zip.writestr(name, source_zip.open(name).read())

    source_zip.close()
    target_zip.close()



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2023 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import FreeCAD as App

from PySide.QtCore import QT_TRANSLATE_NOOP

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtCore, QtGui, QtWidgets

import UtilsAssembly
import Preferences

translate = App.Qt.translate

__title__ = "Assembly Command Create Assembly"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"


class CommandCreateAssembly:
    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Geoassembly",
            "MenuText": QT_TRANSLATE_NOOP("Assembly_CreateAssembly", "New Assembly"),
            "Accel": "A",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Assembly_CreateAssembly",
                "Creates an assembly object in the current document, or in the current active assembly (if any). Limit of one root assembly per file.",
            ),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        if Gui.Control.activeDialog():
            return False

        if Preferences.preferences().GetBool("EnforceOneAssemblyRule", True):
            activeAssembly = UtilsAssembly.activeAssembly()

            if UtilsAssembly.isThereOneRootAssembly() and not activeAssembly:
                return False

        return App.ActiveDocument is not None

    def Activated(self):
        App.setActiveTransaction("New Assembly")

        activeAssembly = UtilsAssembly.activeAssembly()
        Gui.addModule("UtilsAssembly")
        if activeAssembly:
            commands = (
                "activeAssembly = UtilsAssembly.activeAssembly()\n"
                'assembly = activeAssembly.newObject("Assembly::AssemblyObject", "Assembly")\n'
            )
        else:
            commands = (
                'assembly = App.ActiveDocument.addObject("Assembly::AssemblyObject", "Assembly")\n'
            )

        commands = commands + 'assembly.Type = "Assembly"\n'
        commands = commands + 'assembly.newObject("Assembly::JointGroup", "Joints")'

        Gui.doCommand(commands)
        if not activeAssembly:
            Gui.doCommandGui("Gui.ActiveDocument.setEdit(assembly)")

        App.closeActiveTransaction()


class ActivateAssemblyTaskPanel:
    """A basic TaskPanel to select an assembly to activate."""

    def __init__(self, assemblies):
        self.assemblies = assemblies
        self.form = QtWidgets.QWidget()
        self.form.setWindowTitle(translate("Assembly_ActivateAssembly", "Activate Assembly"))

        layout = QtWidgets.QVBoxLayout(self.form)
        label = QtWidgets.QLabel(
            translate("Assembly_ActivateAssembly", "Select an assembly to activate:")
        )
        self.combo = QtWidgets.QComboBox()

        for asm in self.assemblies:
            # Store the user-friendly Label for display, and the internal Name for activation
            self.combo.addItem(asm.Label, asm.Name)

        layout.addWidget(label)
        layout.addWidget(self.combo)

    def accept(self):
        """Called when the user clicks OK."""
        selected_name = self.combo.currentData()
        if selected_name:
            Gui.doCommand(f"Gui.ActiveDocument.setEdit('{selected_name}')")
        return True

    def reject(self):
        """Called when the user clicks Cancel or closes the panel."""
        return True


class CommandActivateAssembly:
    def __init__(self):
        self.task_panel = None

    def GetResources(self):
        return {
            "Pixmap": "Assembly_ActivateAssembly",
            "MenuText": QT_TRANSLATE_NOOP("Assembly_ActivateAssembly", "Activate Assembly"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Assembly_ActivateAssembly", "Sets an assembly as the active one for editing."
            ),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        if Gui.Control.activeDialog() or App.ActiveDocument is None:
            return False

        # Command is only active if no assembly is currently active
        if UtilsAssembly.activeAssembly() is not None:
            return False

        # And if there is at least one assembly in the document to activate
        for obj in App.ActiveDocument.Objects:
            if obj.isDerivedFrom("Assembly::AssemblyObject"):
                return True

        return False

    def Activated(self):
        doc = App.ActiveDocument
        assemblies = [o for o in doc.Objects if o.isDerivedFrom("Assembly::AssemblyObject")]

        if len(assemblies) == 1:
            # If there's only one, activate it directly without showing a dialog
            Gui.doCommand(f"Gui.ActiveDocument.setEdit('{assemblies[0].Name}')")
        elif len(assemblies) > 1:
            # If there are multiple, show a task panel to let the user choose
            self.task_panel = ActivateAssemblyTaskPanel(assemblies)
            Gui.Control.showDialog(self.task_panel)


if App.GuiUp:
    Gui.addCommand("Assembly_CreateAssembly", CommandCreateAssembly())
    Gui.addCommand("Assembly_ActivateAssembly", CommandActivateAssembly())



# ***************************************************************************
# *   Copyright (c) 2023 edi <edi271@a1.net>                                *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the TechDraw HoleShaftFit Task Dialog."""

__title__ = "TechDrawTools.TaskHoleShaftFit"
__author__ = "edi"
__url__ = "https://www.freecad.org"
__version__ = "00.01"
__date__ = "2023/02/07"

import FreeCAD as App
import FreeCADGui as Gui

from functools import partial

import os

translate = App.Qt.translate


class TaskHoleShaftFit:
    def __init__(self, sel):

        loose = translate("TechDraw_HoleShaftFit", "Loose fit")
        snug = translate("TechDraw_HoleShaftFit", "Snug fit")
        press = translate("TechDraw_HoleShaftFit", "Press fit")
        self.isHole = True
        self.sel = sel
        self.holeValues = [
            ["h9", "D10", loose],
            ["h9", "E9", loose],
            ["h9", "F8", loose],
            ["h6", "G7", loose],
            ["c11", "H11", loose],
            ["f7", "H8", loose],
            ["h6", "H7", loose],
            ["h7", "H8", loose],
            ["k6", "H7", snug],
            ["n6", "H7", snug],
            ["r6", "H7", press],
            ["s6", "H7", press],
            ["h6", "K7", snug],
            ["h6", "N7", snug],
            ["h6", "R7", press],
            ["h6", "S7", press],
        ]
        self.shaftValues = [
            ["H11", "c11", loose],
            ["H8", "f7", loose],
            ["H7", "h6", loose],
            ["H8", "h7", loose],
            ["D10", "h9", loose],
            ["E9", "h9", loose],
            ["F8", "h9", loose],
            ["G7", "h6", loose],
            ["K7", "h6", snug],
            ["N7", "h6", snug],
            ["R7", "h6", press],
            ["S7", "h6", press],
            ["H7", "k6", snug],
            ["H7", "n6", snug],
            ["H7", "r6", press],
            ["H7", "s6", press],
        ]

        self._uiPath = App.getHomePath()
        self._uiPath = os.path.join(
            self._uiPath, "Mod/TechDraw/TechDrawTools/Gui/TaskHoleShaftFit.ui"
        )
        self.form = Gui.PySideUic.loadUi(self._uiPath)

        self.form.setWindowTitle(
            translate("TechDraw_HoleShaftFit", "Hole/Shaft Fit ISO 286")
        )

        self.form.rbHoleBase.clicked.connect(partial(self.on_HoleShaftChanged, True))
        self.form.rbShaftBase.clicked.connect(partial(self.on_HoleShaftChanged, False))
        self.form.cbField.currentIndexChanged.connect(self.on_FieldChanged)

        App.setActiveTransaction("Add hole or shaft fit")

    def setHoleFields(self):
        """set hole fields in the combo box"""
        for i in range(self.form.cbField.count()):
            self.form.cbField.removeItem(0)
        for value in self.holeValues:
            self.form.cbField.addItem(value[1])
        self.form.lbBaseField.setText("             " + self.holeValues[0][0] + " /")
        self.form.lbFitType.setText(self.holeValues[0][2])

    def setShaftFields(self):
        """set shaft fields in the combo box"""
        for i in range(self.form.cbField.count()):
            self.form.cbField.removeItem(0)
        for value in self.shaftValues:
            self.form.cbField.addItem(value[1])
        self.form.lbBaseField.setText("             " + self.shaftValues[0][0] + " /")
        self.form.lbFitType.setText(self.shaftValues[0][2])

    def on_HoleShaftChanged(self, isHole):
        """slot: change the used base fit hole/shaft"""
        if isHole:
            self.isHole = isHole
            self.setShaftFields()
        else:
            self.isHole = isHole
            self.setHoleFields()

    def on_FieldChanged(self):
        """slot: change of the desired field"""
        currentIndex = self.form.cbField.currentIndex()
        if self.isHole:
            self.form.lbBaseField.setText(
                "             " + self.shaftValues[currentIndex][0] + " /"
            )
            self.form.lbFitType.setText(self.shaftValues[currentIndex][2])
        else:
            self.form.lbBaseField.setText(
                "             " + self.holeValues[currentIndex][0] + " /"
            )
            self.form.lbFitType.setText(self.holeValues[currentIndex][2])

    def accept(self):
        """slot: OK pressed"""
        currentIndex = self.form.cbField.currentIndex()
        if self.isHole:
            selectedField = self.shaftValues[currentIndex][1]
        else:
            selectedField = self.holeValues[currentIndex][1]
        fieldChar = selectedField[0]
        quality = int(selectedField[1:])
        dim = self.sel[0].Object
        value = dim.getRawValue()
        iso = ISO286()
        iso.calculate(value, fieldChar, quality)
        rangeValues = iso.getValues()
        mainFormat = dim.FormatSpec
        dim.FormatSpec = mainFormat + " " + selectedField
        dim.EqualTolerance = False
        dim.OverTolerance = rangeValues[0]
        dim.UnderTolerance = rangeValues[1]
        if dim.OverTolerance < 0:
            dim.FormatSpecOverTolerance = "(%-0.6w)"
        elif dim.OverTolerance > 0:
            dim.FormatSpecOverTolerance = "(+%-0.6w)"
        else:
            dim.FormatSpecOverTolerance = "( %-0.6w)"
        if dim.UnderTolerance < 0:
            dim.FormatSpecUnderTolerance = "(%-0.6w)"
        elif dim.UnderTolerance > 0:
            dim.FormatSpecUnderTolerance = "(+%-0.6w)"
        else:
            dim.FormatSpecUnderTolerance = "( %-0.6w)"
        Gui.Control.closeDialog()
        App.closeActiveTransaction()

    def reject(self):
        App.closeActiveTransaction(True)
        return True


class ISO286:
    """This class represents a subset of the ISO 286 standard"""

    def getNominalRange(self, measureValue):
        """return index of selected nominal range field, 0 < measureValue < 500 mm"""
        measureRanges = [
            0,
            3,
            6,
            10,
            14,
            18,
            24,
            30,
            40,
            50,
            65,
            80,
            100,
            120,
            140,
            160,
            180,
            200,
            225,
            250,
            280,
            315,
            355,
            400,
            450,
            500,
        ]
        index = 1
        while measureValue > measureRanges[index]:
            index = index + 1
        return index - 1

    def getITValue(self, valueQuality, valueNominalRange):
        """return IT-value  (value of quality in micrometers)"""
        """tables IT6 to IT11 from 0 to 500 mm"""
        IT6 = [
            6,
            8,
            9,
            11,
            11,
            13,
            13,
            16,
            16,
            19,
            19,
            22,
            22,
            25,
            25,
            25,
            29,
            29,
            29,
            32,
            32,
            36,
            36,
            40,
            40,
        ]
        IT7 = [
            10,
            12,
            15,
            18,
            18,
            21,
            21,
            25,
            25,
            30,
            30,
            35,
            35,
            40,
            40,
            40,
            46,
            46,
            46,
            52,
            52,
            57,
            57,
            63,
            63,
        ]
        IT8 = [
            14,
            18,
            22,
            27,
            27,
            33,
            33,
            39,
            39,
            46,
            46,
            54,
            54,
            63,
            63,
            63,
            72,
            72,
            72,
            81,
            81,
            89,
            89,
            97,
            97,
        ]
        IT9 = [
            25,
            30,
            36,
            43,
            43,
            52,
            52,
            62,
            62,
            74,
            74,
            87,
            87,
            100,
            100,
            100,
            115,
            115,
            115,
            130,
            130,
            140,
            140,
            155,
            155,
        ]
        IT10 = [
            40,
            48,
            58,
            70,
            70,
            84,
            84,
            100,
            100,
            120,
            120,
            140,
            140,
            160,
            160,
            160,
            185,
            185,
            185,
            210,
            210,
            230,
            230,
            250,
            250,
        ]
        IT11 = [
            60,
            75,
            90,
            110,
            110,
            130,
            130,
            160,
            160,
            190,
            190,
            220,
            220,
            250,
            250,
            250,
            290,
            290,
            290,
            320,
            320,
            360,
            360,
            400,
            400,
        ]
        qualityTable = [IT6, IT7, IT8, IT9, IT10, IT11]
        return qualityTable[valueQuality - 6][valueNominalRange]

    def getFieldValue(self, fieldCharacter, valueNominalRange):
        """return es or ES value of the field in micrometers"""
        cField = [
            -60,
            -70,
            -80,
            -95,
            -95,
            -110,
            -110,
            -120,
            -130,
            -140,
            -150,
            -170,
            -180,
            -200,
            -210,
            -230,
            -240,
            -260,
            -280,
            -300,
            -330,
            -360,
            -400,
            -440,
            -480,
        ]
        fField = [
            -6,
            -10,
            -13,
            -16,
            -16,
            -20,
            -20,
            -25,
            -25,
            -30,
            -30,
            -36,
            -36,
            -43,
            -43,
            -43,
            -50,
            -50,
            -50,
            -56,
            -56,
            -62,
            -62,
            -68,
            -68,
        ]
        gField = [
            -2,
            -4,
            -5,
            -6,
            -6,
            -7,
            -7,
            -9,
            -9,
            -10,
            -10,
            -12,
            -12,
            -14,
            -14,
            -14,
            -15,
            -15,
            -15,
            -17,
            -17,
            -18,
            -18,
            -20,
            -20,
        ]
        hField = [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ]
        kField = [
            6,
            9,
            10,
            12,
            12,
            15,
            15,
            18,
            18,
            21,
            21,
            25,
            25,
            28,
            28,
            28,
            33,
            33,
            33,
            36,
            36,
            40,
            40,
            45,
            45,
        ]
        nField = [
            10,
            16,
            19,
            23,
            23,
            28,
            28,
            33,
            33,
            39,
            39,
            45,
            45,
            52,
            52,
            60,
            60,
            66,
            66,
            73,
            73,
            80,
            80,
        ]
        rField = [
            16,
            23,
            28,
            34,
            34,
            41,
            41,
            50,
            50,
            60,
            62,
            73,
            76,
            88,
            90,
            93,
            106,
            109,
            113,
            126,
            130,
            144,
            150,
            166,
            172,
        ]
        sField = [
            20,
            27,
            32,
            39,
            39,
            48,
            48,
            59,
            59,
            72,
            78,
            93,
            101,
            117,
            125,
            133,
            151,
            159,
            169,
            190,
            202,
            226,
            244,
            272,
            292,
        ]
        DField = [
            60,
            78,
            98,
            120,
            120,
            149,
            149,
            180,
            180,
            220,
            220,
            260,
            260,
            305,
            305,
            305,
            355,
            355,
            355,
            400,
            400,
            440,
            440,
            480,
            480,
        ]
        EField = [
            39,
            50,
            61,
            75,
            75,
            92,
            92,
            112,
            112,
            134,
            134,
            159,
            159,
            185,
            185,
            185,
            215,
            215,
            215,
            240,
            240,
            265,
            265,
            290,
            290,
        ]
        FField = [
            20,
            28,
            35,
            43,
            43,
            53,
            53,
            64,
            64,
            76,
            76,
            90,
            90,
            106,
            106,
            106,
            122,
            122,
            122,
            137,
            137,
            151,
            151,
            165,
            165,
        ]
        GField = [
            12,
            16,
            20,
            24,
            24,
            28,
            28,
            34,
            34,
            40,
            40,
            47,
            47,
            54,
            54,
            54,
            61,
            61,
            61,
            69,
            69,
            75,
            75,
            83,
            83,
        ]
        HField = [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ]
        KField = [
            0,
            3,
            5,
            6,
            6,
            6,
            6,
            7,
            7,
            9,
            9,
            10,
            10,
            12,
            12,
            12,
            13,
            13,
            13,
            16,
            16,
            17,
            17,
            18,
            18,
        ]
        NField = [
            -4,
            -4,
            -4,
            -5,
            -5,
            -7,
            -7,
            -8,
            -8,
            -9,
            -9,
            -10,
            -10,
            -12,
            -12,
            -12,
            -14,
            -14,
            -14,
            -14,
            -14,
            -16,
            -16,
            -17,
            -17,
        ]
        RField = [
            -10,
            -11,
            -13,
            -16,
            -16,
            -20,
            -20,
            -25,
            -25,
            -30,
            -32,
            -38,
            -41,
            -48,
            -50,
            -53,
            -60,
            -63,
            -67,
            -74,
            -78,
            -87,
            -93,
            -103,
            -109,
        ]
        SField = [
            -14,
            -15,
            -17,
            -21,
            -21,
            -27,
            -27,
            -34,
            -34,
            -42,
            -48,
            -58,
            -66,
            -77,
            -85,
            -93,
            -105,
            -113,
            -123,
            -138,
            -150,
            -169,
            -187,
            -209,
            -229,
        ]
        fieldDict = {
            "c": cField,
            "f": fField,
            "g": gField,
            "h": hField,
            "k": kField,
            "n": nField,
            "r": rField,
            "s": sField,
            "D": DField,
            "E": EField,
            "F": FField,
            "G": GField,
            "H": HField,
            "K": KField,
            "N": NField,
            "R": RField,
            "S": SField,
        }
        return fieldDict[fieldCharacter][valueNominalRange]

    def calculate(self, value, fieldChar, quality):
        """calculate upper and lower field values"""
        self.nominalRange = self.getNominalRange(value)
        self.upperValue = self.getFieldValue(fieldChar, self.nominalRange)
        self.lowerValue = self.upperValue - self.getITValue(quality, self.nominalRange)
        if fieldChar == "H":
            self.upperValue = -self.lowerValue
            self.lowerValue = 0

    def getValues(self):
        """return range values in mm"""
        return (self.upperValue / 1000, self.lowerValue / 1000)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009 Yorik van Havre <yorik@uncreated.net>                        *
# *   (c) 2010 Ken Cline <cline@frii.com>                                   *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create straight Line and Wire objects.

The Line class is used by other Gui Commands that behave in a similar way
like Wire, BSpline, and BezCurve.
"""
## @package gui_lines
# \ingroup draftguitools
# \brief Provides GUI tools to create straight Line and Wire objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import DraftVecUtils
from draftguitools import gui_base_original
from draftguitools import gui_tool_utils
from draftutils import gui_utils
from draftutils import params
from draftutils import utils
from draftutils import todo
from draftutils.messages import _err, _toolmsg
from draftutils.translate import translate


class Line(gui_base_original.Creator):
    """Gui command for the Line tool."""

    def __init__(self, mode="line"):
        super().__init__()
        self.mode = mode

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Line",
            "Accel": "L,I",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Line", "Line"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_Line", "Creates a 2-point line"),
        }

    def Activated(
        self, name=QT_TRANSLATE_NOOP("draft", "Line"), icon="Draft_Line", task_title=None
    ):
        """Execute when the command is called."""
        super().Activated(name)
        if task_title is None:
            title = translate("draft", name)
        else:
            title = task_title
        if self.mode == "wire":
            self.ui.wireUi(title=title, icon=icon)
        elif self.mode == "leader":
            self.ui.wireUi(title=title, icon=icon)
            self.ui.closeButton.hide()
            self.ui.makeFace.hide()
        else:
            self.ui.lineUi(title=title, icon=icon)

        self.obj = self.doc.addObject("Part::Feature", self.featureName)
        gui_utils.format_object(self.obj)
        self.obj.ViewObject.ShowInTree = False

        self.call = self.view.addEventCallback("SoEvent", self.action)
        _toolmsg(translate("draft", "Pick first point"))

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
            return
        if arg["Type"] == "SoLocation2Event":
            self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg)
            gui_tool_utils.redraw3DView()
            return
        if arg["Type"] != "SoMouseButtonEvent":
            return
        if arg["State"] == "UP":
            self.obj.ViewObject.Selectable = True
            return
        if arg["State"] == "DOWN" and arg["Button"] == "BUTTON1":
            # Stop self.obj from being selected to avoid its display in the tree:
            self.obj.ViewObject.Selectable = False
            if arg["Position"] == self.pos:
                self.finish(cont=None)
                return
            if (not self.node) and (not self.support):
                gui_tool_utils.getSupport(arg)
                self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg)
            if self.point:
                self.ui.redraw()
                self.pos = arg["Position"]
                self.node.append(self.point)
                self.drawUpdate(self.point)
                if self.mode == "line" and len(self.node) == 2:
                    self.finish(cont=None, closed=False)
                if len(self.node) > 2:
                    # The wire is closed
                    if (self.point - self.node[0]).Length < utils.tolerance():
                        self.undolast()
                        if len(self.node) > 2:
                            self.finish(cont=None, closed=True)
                        else:
                            self.finish(cont=None, closed=False)

    def finish(self, cont=False, closed=False):
        """Terminate the operation and close the polyline if asked.

        Parameters
        ----------
        cont: bool or None, optional
            Restart (continue) the command if `True`, or if `None` and
            `ui.continueMode` is `True`.
        closed: bool, optional
            Close the line if `True`.
        """
        self.end_callbacks(self.call)
        self.removeTemporaryObject()

        if len(self.node) > 1:
            Gui.addModule("Draft")
            # The command to run is built as a series of text strings
            # to be committed through the `draftutils.todo.ToDo` class.
            if len(self.node) == 2 and params.get_param("UsePartPrimitives"):
                # Insert a Part::Primitive object
                p1 = self.node[0]
                p2 = self.node[-1]

                _cmd = "FreeCAD.ActiveDocument."
                _cmd += 'addObject("Part::Line", "Line")'
                _cmd_list = [
                    "line = " + _cmd,
                    "line.X1 = " + str(p1.x),
                    "line.Y1 = " + str(p1.y),
                    "line.Z1 = " + str(p1.z),
                    "line.X2 = " + str(p2.x),
                    "line.Y2 = " + str(p2.y),
                    "line.Z2 = " + str(p2.z),
                    "Draft.autogroup(line)",
                    "Draft.select(line)",
                    "FreeCAD.ActiveDocument.recompute()",
                ]
                self.commit(translate("draft", "Create Line"), _cmd_list)
            else:
                # Insert a Draft line
                rot, sup, pts, fil = self.getStrings()

                _base = DraftVecUtils.toString(self.node[0])
                _cmd = "Draft.make_wire"
                _cmd += "("
                _cmd += "points, "
                _cmd += "placement=pl, "
                _cmd += "closed=" + str(closed) + ", "
                _cmd += "face=" + fil + ", "
                _cmd += "support=" + sup
                _cmd += ")"
                _cmd_list = [
                    "pl = FreeCAD.Placement()",
                    "pl.Rotation.Q = " + rot,
                    "pl.Base = " + _base,
                    "points = " + pts,
                    "line = " + _cmd,
                    "Draft.autogroup(line)",
                    "FreeCAD.ActiveDocument.recompute()",
                ]
                self.commit(translate("draft", "Create Wire"), _cmd_list)
        super().finish()
        if cont or (cont is None and self.ui and self.ui.continueMode):
            self.Activated()

    def removeTemporaryObject(self):
        """Remove temporary object created."""
        if self.obj:
            try:
                old = self.obj.Name
            except ReferenceError:
                # object already deleted, for some reason
                pass
            else:
                todo.ToDo.delay(self.doc.removeObject, old)
        self.obj = None

    def undolast(self):
        """Undoes last line segment."""
        import Part

        if len(self.node) > 1:
            self.node.pop()
            # last = self.node[-1]
            if self.obj.Shape.Edges:
                edges = self.obj.Shape.Edges
                if len(edges) > 1:
                    newshape = Part.makePolygon(self.node)
                    self.obj.Shape = newshape
                else:
                    self.obj.ViewObject.hide()
                # DNC: report on removal
                # _toolmsg(translate("draft", "Removing last point"))
                _toolmsg(translate("draft", "Pick next point"))
            self.update_hints()

    def drawUpdate(self, point):
        """Draws new line segment."""
        import Part

        if self.planetrack and self.node:
            self.planetrack.set(self.node[-1])
        if len(self.node) == 1:
            _toolmsg(translate("draft", "Pick next point"))
        elif len(self.node) == 2:
            last = self.node[len(self.node) - 2]
            newseg = Part.LineSegment(last, point).toShape()
            self.obj.Shape = newseg
            self.obj.ViewObject.Visibility = True
            if self.mode != "line":
                _toolmsg(translate("draft", "Pick next point"))
        else:
            currentshape = self.obj.Shape.copy()
            last = self.node[len(self.node) - 2]
            if not DraftVecUtils.equals(last, point):
                newseg = Part.LineSegment(last, point).toShape()
                newshape = currentshape.fuse(newseg)
                self.obj.Shape = newshape
            _toolmsg(translate("draft", "Pick next point"))
        self.update_hints()

    def wipe(self):
        """Remove all previous segments and starts from last point."""
        if len(self.node) > 1:
            # self.obj.Shape.nullify()  # For some reason this fails
            self.obj.ViewObject.Visibility = False
            self.node = [self.node[-1]]
            if self.planetrack:
                self.planetrack.set(self.node[0])
            _toolmsg(translate("draft", "Pick next point"))
            self.update_hints()

    def orientWP(self):
        """Orient the working plane."""
        if len(self.node) > 1 and self.obj:
            import DraftGeomUtils

            n = DraftGeomUtils.getNormal(self.obj.Shape)
            if not n:
                n = self.wp.axis
            p = self.node[-1]
            v = self.node[-1].sub(self.node[-2])
            self.wp.align_to_point_and_axis(p, n, upvec=v, _hist_add=False)
            if self.planetrack:
                self.planetrack.set(self.node[-1])

    def numericInput(self, numx, numy, numz):
        """Validate the entry fields in the user interface.

        This function is called by the toolbar or taskpanel interface
        when valid x, y, and z have been entered in the input fields.
        """
        self.point = App.Vector(numx, numy, numz)
        self.node.append(self.point)
        self.drawUpdate(self.point)
        if self.mode == "line" and len(self.node) == 2:
            self.finish(cont=None, closed=False)
        self.ui.setNextFocus()

    def get_hints(self):
        if len(self.node) == 0:
            hints = [
                Gui.InputHint(translate("draft", "%1 pick first point"), Gui.UserInput.MouseLeft)
            ]
        elif self.mode == "line":
            hints = [
                Gui.InputHint(translate("draft", "%1 pick second point"), Gui.UserInput.MouseLeft)
            ]
        elif len(self.node) > 2:
            hints = [
                Gui.InputHint(
                    translate("draft", "%1 pick next point, snap to first point to close"),
                    Gui.UserInput.MouseLeft,
                )
            ]
        else:
            hints = [
                Gui.InputHint(translate("draft", "%1 pick next point"), Gui.UserInput.MouseLeft)
            ]
        return (
            hints
            + gui_tool_utils._get_hint_xyz_constrain()
            + gui_tool_utils._get_hint_mod_constrain()
            + gui_tool_utils._get_hint_mod_snap()
        )


Gui.addCommand("Draft_Line", Line())


class Wire(Line):
    """Gui command for the Wire or Polyline tool.

    It inherits the `Line` class, and calls essentially the same code,
    only this time the `mode` is set to `"wire"`,
    so we are allowed to place more than two points.
    """

    def __init__(self):
        super().__init__(mode="wire")

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Wire",
            "Accel": "P, L",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Wire", "Polyline"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_Wire", "Creates a polyline"),
        }

    def Activated(self):
        """Execute when the command is called."""
        import Part

        # If there is a selection, and this selection contains various
        # two-point lines, their shapes are extracted, and we attempt
        # to join them into a single Wire (polyline),
        # then the old lines are removed.
        if len(Gui.Selection.getSelection()) > 1:
            edges = []
            for o in Gui.Selection.getSelection():
                if utils.get_type(o) != "Wire":
                    edges = []
                    break
                edges.extend(o.Shape.Edges)
            if edges:
                try:
                    w = Part.Wire(Part.__sortEdges__(edges))
                except Exception:
                    _err(translate("draft", "Unable to create a wire " "from the selected objects"))
                else:
                    # Points of the new fused Wire in string form
                    # 'FreeCAD.Vector(x,y,z), FreeCAD.Vector(x1,y1,z1), ...'
                    pts = ", ".join([str(v.Point) for v in w.Vertexes])
                    pts = pts.replace("Vector ", "FreeCAD.Vector")

                    # List of commands to remove the old objects
                    rems = list()
                    for o in Gui.Selection.getSelection():
                        rems.append("FreeCAD.ActiveDocument." 'removeObject("' + o.Name + '")')

                    Gui.addModule("Draft")
                    # The command to run is built as a series of text strings
                    # to be committed through the `draftutils.todo.ToDo` class
                    _cmd = "wire = Draft.make_wire("
                    _cmd += "[" + pts + "], closed=" + str(w.isClosed())
                    _cmd += ")"
                    _cmd_list = [_cmd]
                    _cmd_list.extend(rems)
                    _cmd_list.append("Draft.autogroup(wire)")
                    _cmd_list.append("FreeCAD.ActiveDocument.recompute()")

                    _op_name = translate("draft", "Convert to Wire")
                    todo.ToDo.delayCommit([(_op_name, _cmd_list)])
                    return

        # If there was no selection or the selection was just one object
        # then we proceed with the normal line creation functions,
        # only this time we will be able to input more than two points
        super().Activated(
            name="Polyline", icon="Draft_Wire", task_title=translate("draft", "Polyline")
        )


Gui.addCommand("Draft_Wire", Wire())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import FreeCAD

if FreeCAD.GuiUp:
    import FreeCADGui
    from FreeCADGui import Workbench
else:
    # Provide a dummy Workbench class when GUI is not available
    class Workbench:
        pass


FreeCAD.__unit_test__ += ["TestCAMGui"]


class PathCommandGroup:
    def __init__(self, cmdlist, menu, tooltip=None):
        self.cmdlist = cmdlist
        self.menu = menu
        if tooltip is None:
            self.tooltip = menu
        else:
            self.tooltip = tooltip

    def GetCommands(self):
        return tuple(self.cmdlist)

    def GetResources(self):
        return {"MenuText": self.menu, "ToolTip": self.tooltip}

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False


class CAMWorkbench(Workbench):
    "CAM workbench"

    def __init__(self):
        self.__class__.Icon = FreeCAD.getResourceDir() + "Mod/CAM/Resources/icons/CAMWorkbench.svg"
        self.__class__.MenuText = "CAM"
        self.__class__.ToolTip = "CAM workbench"

    def Initialize(self):
        global PathCommandGroup

        # Add preferences pages - before loading PathGui to properly order pages of Path group
        import Path.Dressup.Gui.Preferences as PathPreferencesPathDressup
        import Path.Tool.assets.ui.preferences as AssetPreferences
        import Path.Main.Gui.PreferencesJob as PathPreferencesPathJob

        translate = FreeCAD.Qt.translate

        # load the builtin modules
        import Path
        import PathScripts
        import PathGui
        from PySide import QtCore, QtGui

        FreeCADGui.addLanguagePath(":/translations")
        FreeCADGui.addIconPath(":/icons")
        import Path.GuiInit

        from Path.Main.Gui import JobCmd as PathJobCmd
        from Path.Main.Gui import SanityCmd as SanityCmd
        from Path.Tool.toolbit.ui import cmd as PathToolBitCmd
        from Path.Tool.library.ui import cmd as PathToolBitLibraryCmd

        from Path.Tool.camassets import cam_assets

        cam_assets.setup()

        # Check if CAM asset migration is needed for version upgrade
        from Path.Tool.migration.migration import CAMAssetMigrator

        migrator = CAMAssetMigrator()
        migrator.check_migration_needed()

        from PySide.QtCore import QT_TRANSLATE_NOOP

        import PathCommands
        import subprocess
        from packaging.version import Version, parse

        FreeCADGui.addPreferencePage(
            PathPreferencesPathJob.JobPreferencesPage,
            QT_TRANSLATE_NOOP("QObject", "CAM"),
        )
        FreeCADGui.addPreferencePage(
            AssetPreferences.AssetPreferencesPage,
            QT_TRANSLATE_NOOP("QObject", "CAM"),
        )
        FreeCADGui.addPreferencePage(
            PathPreferencesPathDressup.DressupPreferencesPage,
            QT_TRANSLATE_NOOP("QObject", "CAM"),
        )

        Path.GuiInit.Startup()

        # build commands list
        projcmdlist = ["CAM_Job", "CAM_Post", "CAM_Sanity"]
        toolcmdlist = ["CAM_Inspect", "CAM_SelectLoop", "CAM_OpActiveToggle"]

        simcmdlist = ["CAM_SimulatorGL", "CAM_Simulator"]
        prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/CAM")
        simLegacy = prefs.GetBool("DefaultSimulatorLegacy", False)
        if simLegacy:
            simcmdlist.reverse()

        prepcmdlist = [
            "CAM_Comment",
            "CAM_Stop",
            "CAM_Custom",
            "CAM_Probe",
        ]
        twodopcmdlist = [
            "CAM_Profile",
            "CAM_Pocket_Shape",
            "CAM_MillFace",
            "CAM_Helix",
            "CAM_Adaptive",
            "CAM_Slot",
        ]
        threedopcmdlist = ["CAM_Pocket3D"]
        engravecmdlist = ["CAM_Engrave", "CAM_Deburr", "CAM_Vcarve"]
        drillingcmdlist = ["CAM_Drilling"]
        modcmdlist = ["CAM_OperationCopy", "CAM_Array", "CAM_SimpleCopy"]
        dressupcmdlist = [
            "CAM_DressupArray",
            "CAM_DressupAxisMap",
            "CAM_DressupPathBoundary",
            "CAM_DressupDogbone",
            "CAM_DressupDragKnife",
            "CAM_DressupLeadInOut",
            "CAM_DressupRampEntry",
            "CAM_DressupTag",
            "CAM_DressupZCorrect",
        ]
        extracmdlist = []
        specialcmdlist = []

        toolcmdlist.extend(PathToolBitLibraryCmd.BarList)
        toolbitcmdlist = PathToolBitLibraryCmd.MenuList

        simcmdgroup = ["CAM_SimTools"]
        FreeCADGui.addCommand(
            "CAM_SimTools",
            PathCommandGroup(
                simcmdlist,
                QT_TRANSLATE_NOOP("CAM_SimTools", "Simulators"),
            ),
        )
        engravecmdgroup = ["CAM_EngraveTools"]
        FreeCADGui.addCommand(
            "CAM_EngraveTools",
            PathCommandGroup(
                engravecmdlist,
                QT_TRANSLATE_NOOP("CAM_EngraveTools", "Engraving Operations"),
            ),
        )
        if Path.Preferences.experimentalFeaturesEnabled():
            drillingcmdlist.append("CAM_Tapping")

        if set(["CAM_Drilling", "CAM_Tapping"]).issubset(drillingcmdlist):
            drillingcmdgroup = ["CAM_DrillingTools"]
            FreeCADGui.addCommand(
                "CAM_DrillingTools",
                PathCommandGroup(
                    drillingcmdlist,
                    QT_TRANSLATE_NOOP("CAM_DrillingTools", "Drilling Operations"),
                ),
            )
        else:
            drillingcmdgroup = drillingcmdlist

        dressupcmdgroup = ["CAM_DressupTools"]
        FreeCADGui.addCommand(
            "CAM_DressupTools",
            PathCommandGroup(
                dressupcmdlist,
                QT_TRANSLATE_NOOP("CAM_DressupTools", "Dressup Operations"),
            ),
        )
        threedcmdgroup = threedopcmdlist
        if Path.Preferences.experimentalFeaturesEnabled():
            prepcmdlist.append("CAM_PathShapeTC")
            extracmdlist.extend(["CAM_Area", "CAM_Area_Workplane"])
            specialcmdlist.append("CAM_ThreadMilling")
            twodopcmdlist.append("CAM_Slot")

        if Path.Preferences.advancedOCLFeaturesEnabled():
            try:
                r = subprocess.run(
                    ["camotics", "--version"], capture_output=True, text=True
                ).stderr.strip()
                v = parse(r)

                if v >= Version("1.2.2"):
                    toolcmdlist.append("CAM_Camotics")
            except (FileNotFoundError, ModuleNotFoundError):
                pass
            except subprocess.CalledProcessError as e:
                print(f"Failed to execute camotics command: {e}")
            except ValueError as ve:
                print(f"Version error: {ve}")
            except Exception as ex:
                print(f"An unexpected error occurred: {ex}")

            try:
                try:
                    import ocl  # pylint: disable=unused-variable
                except ImportError:
                    import opencamlib as ocl
                from Path.Op.Gui import Surface
                from Path.Op.Gui import Waterline

                threedopcmdlist.extend(["CAM_Surface", "CAM_Waterline"])
                threedcmdgroup = ["CAM_3dTools"]
                FreeCADGui.addCommand(
                    "CAM_3dTools",
                    PathCommandGroup(
                        threedopcmdlist,
                        QT_TRANSLATE_NOOP("CAM_3dTools", "3D Operations"),
                    ),
                )
            except ImportError:
                if not Path.Preferences.suppressOpenCamLibWarning():
                    FreeCAD.Console.PrintError("OpenCamLib is not working!\n")

        self.appendToolbar(QT_TRANSLATE_NOOP("Workbench", "Project Setup"), projcmdlist)
        self.appendToolbar(
            QT_TRANSLATE_NOOP("Workbench", "Tool Commands"),
            simcmdgroup + toolcmdlist,
        )
        self.appendToolbar(
            QT_TRANSLATE_NOOP("Workbench", "New Operations"),
            twodopcmdlist + drillingcmdgroup + engravecmdgroup + threedcmdgroup,
        )
        self.appendToolbar(
            QT_TRANSLATE_NOOP("Workbench", "Path Modification"), modcmdlist + dressupcmdgroup
        )
        if extracmdlist:
            self.appendToolbar(QT_TRANSLATE_NOOP("Workbench", "Helpful Tools"), extracmdlist)

        self.appendMenu(
            [QT_TRANSLATE_NOOP("Workbench", "&CAM")],
            projcmdlist
            + ["CAM_ExportTemplate", "Separator"]
            + simcmdlist
            + toolcmdlist
            + toolbitcmdlist
            + ["Separator"]
            + twodopcmdlist
            + drillingcmdlist
            + engravecmdlist
            + ["Separator"]
            + threedopcmdlist
            + ["Separator"],
        )
        self.appendMenu(
            [
                QT_TRANSLATE_NOOP("Workbench", "&CAM"),
                QT_TRANSLATE_NOOP("Workbench", "Path Dressup"),
            ],
            dressupcmdlist,
        )
        self.appendMenu(
            [
                QT_TRANSLATE_NOOP("Workbench", "&CAM"),
                QT_TRANSLATE_NOOP("Workbench", "Supplemental Commands"),
            ],
            prepcmdlist,
        )
        self.appendMenu(
            [
                QT_TRANSLATE_NOOP("Workbench", "&CAM"),
                QT_TRANSLATE_NOOP("Workbench", "Path Modification"),
            ],
            modcmdlist,
        )
        if specialcmdlist:
            self.appendMenu(
                [
                    QT_TRANSLATE_NOOP("Workbench", "&CAM"),
                    QT_TRANSLATE_NOOP("Workbench", "Specialty Operations"),
                ],
                specialcmdlist,
            )
        if extracmdlist:
            self.appendMenu([QT_TRANSLATE_NOOP("Workbench", "&CAM")], extracmdlist)

        self.appendMenu([QT_TRANSLATE_NOOP("Workbench", "&CAM")], ["Separator"])
        self.appendMenu(
            [
                QT_TRANSLATE_NOOP("Workbench", "&CAM"),
                QT_TRANSLATE_NOOP("Workbench", "Utils"),
            ],
            ["CAM_PropertyBag"],
        )

        self.dressupcmds = dressupcmdlist

        curveAccuracy = Path.Preferences.defaultLibAreaCurveAccuracy()
        if curveAccuracy:
            Path.Area.setDefaultParams(Accuracy=curveAccuracy)

        # keep this one the last entry in the preferences
        import Path.Base.Gui.PreferencesAdvanced as PathPreferencesAdvanced
        from Path.Preferences import preferences

        FreeCADGui.addPreferencePage(
            PathPreferencesAdvanced.AdvancedPreferencesPage,
            QT_TRANSLATE_NOOP("QObject", "CAM"),
        )
        Log("Loading CAM workbench... done\n")

    def GetClassName(self):
        return "Gui::PythonWorkbench"

    def Activated(self):
        # update the translation engine
        FreeCADGui.updateLocale()
        # Msg("CAM workbench activated\n")

    def Deactivated(self):
        # Msg("CAM workbench deactivated\n")
        pass

    def ContextMenu(self, recipient):
        import PathScripts

        menuAppended = False
        if len(FreeCADGui.Selection.getSelection()) == 1:
            obj = FreeCADGui.Selection.getSelection()[0]
            if obj.isDerivedFrom("Path::Feature"):
                self.appendContextMenu("", "Separator")
                self.appendContextMenu("", ["CAM_Inspect"])
                selectedName = obj.Name
                if "Remote" in selectedName:
                    self.appendContextMenu("", ["Refresh_Path"])
                if "Job" in selectedName:
                    self.appendContextMenu("", ["CAM_ExportTemplate"] + self.toolbitctxmenu)
                menuAppended = True
            if isinstance(obj.Proxy, Path.Op.Base.ObjectOp):
                self.appendContextMenu("", ["CAM_OperationCopy", "CAM_OpActiveToggle"])
                if hasattr(obj, "StartPoint"):
                    self.appendContextMenu("", ["CAM_SetStartPoint"])
                menuAppended = True
            if obj.isDerivedFrom("Path::Feature"):
                if (
                    "Profile" in selectedName
                    or "Contour" in selectedName
                    or "Dressup" in selectedName
                    or "Pocket" in selectedName
                ):
                    self.appendContextMenu("", "Separator")
                    # self.appendContextMenu("", ["Set_StartPoint"])
                    # self.appendContextMenu("", ["Set_EndPoint"])
                    for cmd in self.dressupcmds:
                        self.appendContextMenu("Dressup", [cmd])
                    menuAppended = True
            if isinstance(obj.Proxy, Path.Tool.ToolBit):
                self.appendContextMenu("", ["CAM_ToolBitSave", "CAM_ToolBitSaveAs"])
                menuAppended = True
        if menuAppended:
            self.appendContextMenu("", "Separator")


Gui.addWorkbench(CAMWorkbench())

FreeCAD.addImportType(
    "GCode (*.nc *.NC *.gc *.GC *.ncc *.NCC *.ngc *.NGC *.cnc *.CNC *.tap *.TAP *.gcode *.GCODE)",
    "PathGui",
)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 FreeCAD Team                                       *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENSE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Unit tests for the ToolBitListWidget."""

from typing import cast
import unittest
from Path.Tool.toolbit import ToolBit
from Path.Tool.toolbit.ui.toollist import ToolBitListWidget, ToolBitUriRole
from Path.Tool.toolbit.ui.tablecell import TwoLineTableCell
from .PathTestUtils import PathTestWithAssets  # Import the base test class


class TestToolBitListWidget(PathTestWithAssets):
    """Tests for ToolBitListWidget using real assets."""

    def setUp(self):
        super().setUp()  # Call the base class setUp to initialize assets
        self.widget = ToolBitListWidget()

    def test_add_toolbit(self):
        # Get a real ToolBit asset
        toolbit = cast(ToolBit, self.assets.get("toolbit://5mm_Endmill"))
        tool_no = 1

        self.widget.add_toolbit(toolbit, str(tool_no))

        self.assertEqual(self.widget.count(), 1)
        item = self.widget.item(0)
        self.assertIsNotNone(item)

        cell_widget = self.widget.itemWidget(item)
        self.assertIsInstance(cell_widget, TwoLineTableCell)  # Check against real class

        # Verify cell widget properties are set correctly
        self.assertEqual(cell_widget.tool_no, str(tool_no))
        self.assertEqual(cell_widget.upper_text, toolbit.label)
        # Assuming the 5mm_Endmill asset has a shape named 'Endmill'
        normalized_lower_text = cell_widget.lower_text.replace(",00 ", ".00 ")
        self.assertEqual(normalized_lower_text, "5.00 mm 4-flute endmill, 30.00 mm cutting edge")

        # Verify URI is stored in item data
        stored_uri = item.data(ToolBitUriRole)
        self.assertEqual(stored_uri, str(toolbit.get_uri()))

    def test_clear_list(self):
        # Add some real items first
        toolbit1 = cast(ToolBit, self.assets.get("toolbit://5mm_Endmill"))
        toolbit2 = cast(ToolBit, self.assets.get("toolbit://slittingsaw"))
        self.widget.add_toolbit(toolbit1, 1)
        self.widget.add_toolbit(toolbit2, 2)
        self.assertEqual(self.widget.count(), 2)

        self.widget.clear_list()
        self.assertEqual(self.widget.count(), 0)

    def test_apply_filter(self):
        # Add items with distinct text for filtering
        toolbit1 = cast(ToolBit, self.assets.get("toolbit://5mm_Endmill"))
        toolbit2 = cast(ToolBit, self.assets.get("toolbit://slittingsaw"))
        toolbit3 = cast(ToolBit, self.assets.get("toolbit://probe"))

        self.widget.add_toolbit(toolbit1, 1)
        self.widget.add_toolbit(toolbit2, 2)
        self.widget.add_toolbit(toolbit3, 3)

        items = [self.widget.item(i) for i in range(self.widget.count())]
        cells = [self.widget.itemWidget(item) for item in items]

        # Test filter "Endmill"
        self.widget.apply_filter("Endmill")

        self.assertFalse(items[0].isHidden())  # 5mm Endmill
        self.assertTrue(items[1].isHidden())  # slittingsaw
        self.assertTrue(items[2].isHidden())  # probe

        # Verify highlight was called on all cells
        for cell in cells:
            self.assertEqual(cell.search_highlight, "Endmill")

        # Test filter "Ballnose"
        self.widget.apply_filter("Ballnose")

        self.assertTrue(items[0].isHidden())  # 5mm Endmill
        self.assertTrue(items[1].isHidden())  # slittingsaw
        self.assertTrue(items[2].isHidden())  # probe

        # Verify highlight was called again
        for cell in cells:
            self.assertEqual(cell.search_highlight, "Ballnose")

        # Test filter "3mm"
        self.widget.apply_filter("3mm")

        self.assertTrue(items[0].isHidden())  # 5mm Endmill
        self.assertTrue(items[1].isHidden())  # slittingsaw
        self.assertTrue(items[2].isHidden())  # probe

        # Verify highlight was called again
        for cell in cells:
            self.assertEqual(cell.search_highlight, "3mm")

    def test_get_selected_toolbit_uri(self):
        toolbit1 = cast(ToolBit, self.assets.get("toolbit://5mm_Endmill"))
        toolbit2 = cast(ToolBit, self.assets.get("toolbit://slittingsaw"))

        self.widget.add_toolbit(toolbit1, 1)
        self.widget.add_toolbit(toolbit2, 2)

        # No selection initially
        self.assertIsNone(self.widget.get_selected_toolbit_uri())

        # Select the first item
        self.widget.setCurrentItem(self.widget.item(0))
        self.assertEqual(self.widget.get_selected_toolbit_uri(), str(toolbit1.get_uri()))

        # Select the second item
        self.widget.setCurrentItem(self.widget.item(1))
        self.assertEqual(self.widget.get_selected_toolbit_uri(), str(toolbit2.get_uri()))

        # Clear selection (simulate by setting current item to None)
        self.widget.setCurrentItem(None)
        self.assertIsNone(self.widget.get_selected_toolbit_uri())


if __name__ == "__main__":
    unittest.main()



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

import Part
from BOPTools import SplitFeatures

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Constraint Contact Shell Shell",
        "meshtype": "face",
        "meshelement": "Tria3",
        "constraints": ["displacement", "force", "contact"],
        "solvers": ["ccxtools"],
        "material": "solid",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.constraint_contact_shell_shell import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=42228
based on https://forum.freecad.org/viewtopic.php?f=18&t=42228#p359488

contact example shell to shell elements

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects
    # TODO: turn circle of upper tube to have the line on the other side
    # make a boolean fragment of them to be sure there is a mesh point on remesh
    # but as long as we do not remesh it works without the boolean fragment too

    # tubes
    tube_radius = 25
    tube_length = 500
    sh_lower_circle = Part.Wire(Part.makeCircle(tube_radius))
    sh_lower_tube = sh_lower_circle.extrude(FreeCAD.Vector(0, 0, tube_length))
    sh_lower_tube.reverse()
    lower_tube = doc.addObject("Part::Feature", "Lower_tube")
    lower_tube.Shape = sh_lower_tube

    sh_upper_circle = Part.Wire(Part.makeCircle(tube_radius))
    sh_upper_tube = sh_upper_circle.extrude(FreeCAD.Vector(0, 0, tube_length))
    sh_upper_tube.reverse()
    upper_tube = doc.addObject("Part::Feature", "Upper_tube")
    upper_tube.Shape = sh_upper_tube
    upper_tube.Placement = FreeCAD.Placement(
        FreeCAD.Vector(-25, 51, 475),
        FreeCAD.Rotation(90, 0, 90),
        FreeCAD.Vector(0, 0, 0),
    )

    # point for load
    v_force_pt = FreeCAD.Vector(0, 76, 475)
    sh_force_point = Part.Vertex(v_force_pt)
    force_point = doc.addObject("Part::Feature", "Load_place_point")
    force_point.Shape = sh_force_point
    if FreeCAD.GuiUp:
        force_point.ViewObject.PointSize = 10.0
        force_point.ViewObject.PointColor = (1.0, 0.0, 0.0)

    # boolean fragment of upper tubo and force point
    boolfrag = SplitFeatures.makeBooleanFragments(name="BooleanFragments")
    boolfrag.Objects = [upper_tube, force_point]
    if FreeCAD.GuiUp:
        upper_tube.ViewObject.hide()

    # compound out of bool frag and lower tube
    geom_obj = doc.addObject("Part::Compound", "AllGeomCompound")
    geom_obj.Links = [boolfrag, lower_tube]
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # line for load direction
    sh_load_line = Part.makeLine(v_force_pt, FreeCAD.Vector(0, 150, 475))
    load_line = doc.addObject("Part::Feature", "Load_direction_line")
    load_line.Shape = sh_load_line
    doc.recompute()
    if FreeCAD.GuiUp:
        load_line.ViewObject.LineWidth = 5.0
        load_line.ViewObject.LineColor = (1.0, 0.0, 0.0)

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.AnalysisType = "static"
        solver_obj.BeamShellResultOutput3D = True
        solver_obj.GeometricalNonlinearity = "linear"  # really?
        # TODO iterations parameter !!!
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
        solver_obj.SplitInputWriter = False
    analysis.addObject(solver_obj)

    # shell thickness
    shell_thick = ObjectsFem.makeElementGeometry2D(doc, 0.5, "ShellThickness")
    analysis.addObject(shell_thick)

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "MechanicalMaterial")
    mat = material_obj.Material
    mat["Name"] = "AlCuMgPb"
    mat["YoungsModulus"] = "72000 MPa"
    mat["PoissonRatio"] = "0.30"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # constraint displacement
    con_displacement = ObjectsFem.makeConstraintDisplacement(doc, "ConstraintDisplacement")
    con_displacement.References = [
        (lower_tube, "Edge2"),
        (upper_tube, "Edge3"),
    ]
    con_displacement.xFree = False
    con_displacement.yFree = False
    con_displacement.zFree = False
    analysis.addObject(con_displacement)

    # constraint force
    con_force = ObjectsFem.makeConstraintForce(doc, "ConstraintForce")
    # TODO use point of tube boolean fragment
    con_force.References = [(force_point, "Vertex1")]
    con_force.Force = "5000.0 N"
    con_force.Direction = (load_line, ["Edge1"])
    con_force.Reversed = True
    analysis.addObject(con_force)

    # constraint contact
    con_contact = ObjectsFem.makeConstraintContact(doc, "ConstraintContact")
    con_contact.References = [
        (lower_tube, "Face1"),
        (upper_tube, "Face1"),
    ]
    con_contact.Friction = False
    con_contact.Slope = "1000000.0 GPa/m"
    analysis.addObject(con_contact)

    # mesh
    from .meshes.mesh_contact_tube_tube_tria3 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False

    doc.recompute()
    return doc



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint centrif document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package constraint_centrif
#  \ingroup FEM
#  \brief constraint centrif object

from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper


class ConstraintCentrif(base_fempythonobject.BaseFemPythonObject):
    """
    The ConstraintCentrif object
    """

    Type = "Fem::ConstraintCentrif"

    def __init__(self, obj):
        super().__init__(obj)

        for prop in self._get_properties():
            prop.add_to_object(obj)

    def _get_properties(self):
        prop = []

        prop.append(
            _PropHelper(
                type="App::PropertyFrequency",
                name="RotationFrequency",
                group="Constraint Centrif",
                doc="Set rotation frequency",
                value="0 1/s",
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLinkSubListGlobal",
                name="RotationAxis",
                group="Constraint Centrif",
                doc="Set line as axis of rotation",
                value=[],
            )
        )

        return prop

    def onDocumentRestored(self, obj):
        # update old project with new properties
        for prop in self._get_properties():
            if prop.name == "RotationAxis":
                # change RotationAxis to App::PropertyLinkSubListGlobal
                prop.handle_change_type(obj, old_type="App::PropertyLinkSubList")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM Trash command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_Trash:

    def GetResources(self):
        return {
            "Pixmap": "BIM_Trash",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Trash", "Move to Trash"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_Trash", "Moves the selected objects to the trash folder"
            ),
            "Accel": "Shift+Del",
        }

    def Activated(self):
        if FreeCADGui.Selection.getSelection():
            trash = FreeCAD.ActiveDocument.getObject("Trash")
            if not trash or not trash.isDerivedFrom("App::DocumentObjectGroup"):
                trash = FreeCAD.ActiveDocument.addObject("App::DocumentObjectGroup", "Trash")
                trash.Label = translate("BIM", "Trash")
            for obj in FreeCADGui.Selection.getSelection():
                trash.addObject(obj)
                # check for parents still there
                for par in obj.InList:
                    if (par != trash) and hasattr(par, "Group"):
                        if obj in par.Group:
                            if hasattr(par, "removeObject"):
                                par.removeObject(obj)
                            else:
                                g = par.Group
                                g.remove(obj)
                                par.Group = g
                obj.ViewObject.hide()

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


FreeCADGui.addCommand("BIM_Trash", BIM_Trash())


class BIM_EmptyTrash:
    def GetResources(self):
        return {
            "Pixmap": "BIM_Trash",
            "MenuText": QT_TRANSLATE_NOOP("BIM_EmptyTrash", "Empty Trash"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_EmptyTrash",
                "Deletes from the trash bin all objects that are not used by any other",
            ),
        }

    def Activated(self):
        trash = FreeCAD.ActiveDocument.getObject("Trash")
        if trash and trash.isDerivedFrom("App::DocumentObjectGroup"):
            deletelist = []
            for obj in trash.Group:
                if (len(obj.InList) == 1) and (obj.InList[0] == trash):
                    deletelist.append(obj.Name)
                    deletelist.extend(self.getDeletableChildren(obj))
            if deletelist:
                FreeCAD.ActiveDocument.openTransaction("Empty Trash")
                for name in deletelist:
                    FreeCAD.ActiveDocument.removeObject(name)
                FreeCAD.ActiveDocument.commitTransaction()

    def getDeletableChildren(self, obj):
        deletelist = []
        for child in obj.OutList:
            if (len(child.InList) == 1) and (child.InList[0] == obj):
                deletelist.append(child.Name)
                deletelist.extend(self.getDeletableChildren(child))
        return deletelist
